/*
 * ===========================================
 * Java Pdf Extraction Decoding Access Library
 * ===========================================
 *
 * Project Info:  http://www.jpedal.org
 * Project Lead:  Mark Stephens (mark@idrsolutions.com)
 *
 * (C) Copyright 2002, IDRsolutions and Contributors.
 *
 * ---------------
 * ExtractTextAsWordlist.java
 * ---------------
 * (C) Copyright 2002, by IDRsolutions and Contributors.
 *
 * Original Author:  Mark Stephens (mark@idrsolutions.com)
 * Contributor(s):
 */
package org.jpedal.examples.text.extractheadlines;

import java.io.File;

import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import org.jpedal.PdfDecoder;
import org.jpedal.examples.text.ExtractTextInRectangle;
import org.jpedal.exception.PdfException;
import org.jpedal.exception.PdfSecurityException;
import org.jpedal.grouping.PdfGroupingAlgorithms;
import org.jpedal.objects.PdfPageData;
import org.jpedal.utils.LogWriter;

/**
 * This example was written to show extraction from a page location of
 * repetitive information (ie Section)<br>Scope:<b>(1.4/1.5 Ent only)</b>
 * 
 */
public class ExtractDate extends ExtractTextInRectangle{
	
	/**debug flag to exit if no section found*/
	final private boolean debug=false;
	
	/**holds configuration data*/
	DateConfiguration config;
	
	/**default value*/
	String[] dateTokens=null;
	
	private int[] x1,x2,y1,y2;
	
	private String day,date,month,year;
	
	public String getDay() {
		return day;
	}

	public String getDate() {
		return date;
	}

	public String getMonth() {
		return month;
	}

	public String getYear() {
		return year;
	}

	/**
	 * extract date using tags
	 */
	private String extractDate(String extractedText) {
		
		String pageNumber, date = "", currentToken;
		try {
			if(showMessages)
				System.out.println(extractedText);
			
			if(extractedText==null)
				return null;
			
			Map dates=new HashMap();
			
			int dateTokenCount=dateTokens.length;
			for(int i=0;i<dateTokenCount;i++)
				dates.put(dateTokens[i],"x");
			
			pageNumber = null;
			currentToken = null;
			
			//cycle through to get value
			StringTokenizer tokens=new StringTokenizer(extractedText,"<>");
			
			while(tokens.hasMoreTokens()){
				
				//exit if both found
				if((date!=null)&&(pageNumber!=null))
					break;
				currentToken=tokens.nextToken();
				
				
				//now look for match for page and date
				if((dates.get(currentToken)!=null)){
					String font = currentToken;
					currentToken=tokens.nextToken();
					
					//see if number and ignore if so
					boolean isNumber=false;
					
					if((!isNumber)&&(currentToken.length()>2)){
						StringBuffer dateName=new StringBuffer();
						while(tokens.hasMoreTokens()&&(!currentToken.equals("/font"))){
							if(currentToken.indexOf("SpaceC")!=-1)
								dateName.append(' ');
							else
								dateName.append(currentToken);
							currentToken=tokens.nextToken();
						}
						
						//if(date!=null)
						date=date+" "+dateName.toString().trim();

					}
				}
				
			}
		} catch (RuntimeException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return date;
	}
	
	public static void main(String[] args) {
		
		File file = new File("c:/");
		
		File[] files = file.listFiles();
		
		for(int i=0;i<files.length;i++){
			File currentFile = files[i];
			if(!currentFile.isDirectory()){
				ExtractDate test = new ExtractDate(currentFile.getAbsolutePath(),"config/TIM/");
				
				System.out.println(currentFile.getName());
				System.out.println("\tday = "+test.getDay());
				System.out.println("\tdate = "+test.getDate());
				System.out.println("\tmonth = "+test.getMonth());
				System.out.println("\tyear = "+test.getYear());
			}
		}
		
//		ExtractDate test = new ExtractDate("c:/Documents and Settings/simon/Desktop/times files/TIM061GN10.pdf","config/TIM/");
//		
//		System.out.println(test.getDate());
		
	}
	
	/**example method to open a file and extract the raw text*/
	public ExtractDate(String file_name, String configDir) {
		
		try{
			showMessages=false;
			config=new DateConfiguration(configDir);
			
			/**
			 * read XML tags to look for
			 */
			
			date=config.getValue("default_date");
			
			//get number of tags and init store
			int tagCount=Integer.parseInt(config.getValue("xmlCount"));
			dateTokens=new String[tagCount];
			
			//read in xml tags
			for(int j=0;j<tagCount;j++){
				dateTokens[j]=config.getValue("xmlTag_"+j);
				
				if(showMessages)
					System.out.println(dateTokens[j]);
			}
			
			
			/**
			 * read location values
			 */
			//get number of tags and init store
			tagCount=Integer.parseInt(config.getValue("locationCount"));
			
			x1=new int[tagCount];
			x2=new int[tagCount];
			y1=new int[tagCount];
			y2=new int[tagCount];
			
			//read values
			String key="locTag";
			String[] coords={"x1","y1","x2","y2"};
			
			for(int i=0;i<tagCount;i++){
				
				for(int coord=0;coord<4;coord++){
					
					String currentKey=key+"_"+i+"_"+coords[coord];
					String value=config.getValue(currentKey);
					int numberValue=Integer.parseInt(value);
					
					//set values
					switch(coord){
					case 0:
						x1[i]=numberValue;
						break;
					case 1:
						y1[i]=numberValue;
						break;
					case 2:
						x2[i]=numberValue;
						break;
					case 3:
						y2[i]=numberValue;
						break;
					}
				}
			}
			
			/**/
			
			//check output dir has separator
			if (user_dir.endsWith(separator) == false)
				user_dir = user_dir + separator;
			
			//create a directory if it doesn't exist
			File output_path = new File(outputDir);
			if (output_path.exists() == false)
				output_path.mkdirs();
			
			/**
			 * if file name ends pdf, do the file otherwise 
			 * do every pdf file in the directory. We already know file or
			 * directory exists so no need to check that, but we do need to
			 * check its a directory
			 */
			if (file_name.toLowerCase().endsWith(".pdf")) {
				decodeFile("",file_name);
				
			} else {
				
				/**
				 * get list of files and check directory
				 */
				
				String[] files = null;
				File inputFiles = null;
				
				/**make sure name ends with a deliminator for correct path later*/
				if (!file_name.endsWith(separator))
					file_name = file_name + separator;
				
				try {
					inputFiles = new File(file_name);
					
					if (!inputFiles.isDirectory()) {
						System.err.println(
								file_name + " is not a directory. Exiting program");
					}
					files = inputFiles.list();
				} catch (Exception ee) {
					LogWriter.writeLog(
							"Exception trying to access file " + ee.getMessage());
				}
				
				/**now work through all pdf files*/
				long fileCount = files.length;
				
				for (int i = 0; i < fileCount; i++) {
					if(showMessages)
						System.out.println(i + "/ " + fileCount + " " + files[i]);
					
					if (files[i].toLowerCase().endsWith(".pdf")) {
						if(showMessages)
							System.out.println(file_name + files[i]);
						
						decodeFile(file_name,files[i]);
						
					}
				}
			}
		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	
	/**
	 * routine to decode a file
	 */
	protected void decodeFile(String path,String name) {
		
		try {
			String file_name=path+name;
			
			//date from page
			String date=null;
			
			//PdfDecoder returns a PdfException if there is a problem
			try {
				decodePdf = new PdfDecoder(false);
				decodePdf.setExtractionMode(PdfDecoder.TEXT); //extract just text
				decodePdf.init(true);
				//make sure widths in data CRITICAL if we want to split lines correctly!!
				decodePdf.openPdfFile(file_name);
				
				if(showMessages)
					System.out.println("file_name="+file_name);
				
			} catch (PdfSecurityException se) {
				System.err.println("Security Exception " + se + " in pdf code for text extraction on file "+decodePdf.getObjectStore().getCurrentFilename());
				//e.printStackTrace();
			} catch (PdfException se) {
				System.err.println("Pdf Exception " + se + " in pdf code for text extraction on file "+decodePdf.getObjectStore().getCurrentFilename());
				//e.printStackTrace();
			} catch (Exception e) {
				System.err.println("Exception " + e + " in pdf code for text extraction on file "+decodePdf.getObjectStore().getCurrentFilename());
				e.printStackTrace();
			}
			
			/**
			 * extract data from pdf (if allowed). 
			 */
			if ((decodePdf.isEncrypted()&&(!decodePdf.isPasswordSupplied())) && (!decodePdf.isExtractionAllowed())) {
				if(showMessages){
					System.out.println("Encrypted settings");
					System.out.println(
					"Please look at SimpleViewer for code sample to handle such files");
					System.out.println("Or get support/consultancy");
				}
			} else {
				//page range
				int start = 1, end = decodePdf.getPageCount();
				
				//number of possible values to match
				int possSetsCoordinates=x2.length;
				
				//reset n to null
				date=null;
				
				/**
				 * extract data from pdf
				 */
				try {
					for (int page = start; page < end + 1; page++) { //read pages
						
						//decode the page
						decodePdf.decodePage(page);
						
						/**
						 * scan possible page locations for section title
						 */
						for(int coordSet=0;coordSet<possSetsCoordinates;coordSet++){
							
							//SET co-ordinates
							int x1=this.x1[coordSet];
							int x2=this.x2[coordSet];
							int y1=this.y1[coordSet];
							int y2=this.y2[coordSet];
							
							if(showMessages)
								System.out.println("Using ("+x1+","+y1+") ("+x2+","+y2+")");
							
							/** create a grouping object to apply grouping to data*/
							PdfGroupingAlgorithms currentGrouping =decodePdf.getGroupingObject();
							
							/**use whole page size for  demo - get data from PageData object*/
							PdfPageData currentPageData = decodePdf.getPdfPageData();
							
							/**The call to extract the text*/
							text =null;
							
							try{
								text =currentGrouping.extractTextInRectangle(x1,y1,x2,y2,page,false,true);
							} catch (PdfException e) {
								decodePdf.closePdfFile();
								System.err.println("Exception " + e.getMessage()+" in file "+decodePdf.getObjectStore().fullFileName);
								e.printStackTrace();
							}
							
							if (text == null) {
								if(showMessages)
									System.out.println("No text found");
								
							}else{
								date=extractDate(text);
								
								if(date!=null)
									coordSet=possSetsCoordinates;
							}
						}
						
						//remove data once written out
						decodePdf.flushObjectValues(false);
						
						if(date != null && !date.equals(config.getValue("default_date"))){
							StringTokenizer st = new StringTokenizer(date," ");
							
							this.day = st.nextToken();
							this.month= st.nextToken();
							this.date = st.nextToken();
							this.year = st.nextToken();
							
						}
						
						if(showMessages)
							System.out.println("----");
						
					}
				} catch (Exception e) {
					decodePdf.closePdfFile();
					LogWriter.writeLog("No valid date for "+decodePdf.getObjectStore().getCurrentFilename());
					//e.printStackTrace();
					//System.out.println(decodePdf.getObjectStore().getCurrentFilename());
				}
				
				/**
				 * flush data structures - not strictly required but included
				 * as example
				 */
				decodePdf.flushObjectValues(true); //flush any text data read
				
				/**tell user*/
				if(showMessages)
					System.out.println("Text read");
				
			}
			
			/**close the pdf file*/
			decodePdf.closePdfFile();
		} catch (RuntimeException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
}
