package org.jpedal.objects.acroforms;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Insets;
import java.awt.image.BufferedImage;
import java.awt.image.ByteLookupTable;
import java.awt.image.LookupOp;
import java.util.HashMap;
import java.util.Map;

import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPasswordField;
import javax.swing.JRadioButton;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.border.BevelBorder;
import javax.swing.border.Border;

import org.jpedal.gui.ShowGUIMessage;
import org.jpedal.utils.LogWriter;
import org.jpedal.utils.Strip;

/**
 * default set of methds to create Form fields using standard Swing components.
 * Can be replaced with your own set if required.
 */
public class DefaultFormFactory implements FormFactory{

    final private static boolean debugUnimplemented = false;//to show unimplemented parts
    
    /**
     * shows the icon images as they are being applied to the component
     */
    private static boolean showIconImages=false;
    
    private AcroRenderer acrorend;
    
    //prints all debugging information
    private boolean printAllouts=false;
    //used when only one method needs debugging,
    //see create(Button or choice or text) first line /**/
    private boolean printouts=printAllouts;
    
    private ActionHandler formsActionHandler;
    
    private DefaultFormFactory(){}
    
    /**
     * allows access to rennderer variables
     * @param actionHandler 
     */
    public DefaultFormFactory(AcroRenderer acroRenderer, ActionHandler actionHandler){
        acrorend = acroRenderer;
        formsActionHandler = actionHandler;
    }
    
    public void reset(AcroRenderer acroRenderer, ActionHandler actionHandler){
    	acrorend = acroRenderer;
        formsActionHandler = actionHandler;
    }
    
    /**
     * setup and return the ComboBox field specified in the FormObject
     */
    public Component comboBox(final FormObject form) {
        if(printouts)
        	System.out.println("combobox");
        
        //populate items array with list from Opt
        String[] items = form.itemsList;
        JComboBox comboBox;
        if(items==null)
        	comboBox = new JComboBox();
        else
        	comboBox = new JComboBox(items);
        comboBox.setOpaque(false);
        
        //get and set currently selected value
        String textValue = form.selectedItem;
        comboBox.setSelectedItem(textValue);
        if(printouts)
        	System.out.println("currently selected value="+textValue);
        
        if(form.flags[18]){//FormStream.EDIT
        	if(printouts)
        		System.out.println("drop list and an editable text box");
        	comboBox.setEditable(true);
        	
        }else{//is not editable
        	if(printouts)
        		System.out.println("only a drop list");
        	comboBox.setEditable(false);
        }
        
       	comboBox.setFont(form.textFont);
       	comboBox.setForeground(form.textColor);
       	
		Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
        comboBox.setBorder(newBorder);
        if(printouts){
        	System.out.println("borderStyle="+newBorder);
        }
        
        if(form.backgroundColor!=null){
            comboBox.setBackground(form.backgroundColor);
            comboBox.setOpaque(true);
        }
        
        boolean[] fieldCharacteristic=form.characteristic;
        Map enteredAction = form.enteredAction;
        Map exitedAction = form.exitedAction;
        Map activateAction = form.activateAction;
        
        setupMouseListener(comboBox,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            comboBox.setEditable(false);
//          retComponent.setForeground(Color.magenta);
//          retComponent.setBackground(Color.magenta);
            if(printouts)
                System.out.println("READONLY="+comboBox);
        }
        
        return comboBox;
    }


    /**
     * setup the events for currentComp, from the specified parameters
     */
    private void setupMouseListener(final Component currentComp, boolean[] fieldCharacteristic,
    		final Map enteredAction,final Map exitedAction, final Map activateAction) {
    	/* bit 1 is index 0 in boolean[]
    	 * 1 = invisible
    	 * 2 = hidden - dont display or print
    	 * 3 = print - print if set, dont if not
    	 * 4 = nozoom
    	 * 5= norotate
    	 * 6= noview
    	 * 7 = read only (ignored by wiget)
    	 * 8 = locked
    	 * 9 = togglenoview
    	 */
    	if(fieldCharacteristic[0] || fieldCharacteristic[1] || fieldCharacteristic[5]){
    		currentComp.setVisible(false);
    	}
    	
    	if(fieldCharacteristic[8]){ 
    		/*set visible on mouse over*/
    		currentComp.addMouseListener(formsActionHandler.setupCommand("togglenoview",acrorend));
    	}

    	/*entered Action*/
    	if(enteredAction!=null){
    		if(currentComp instanceof JComboBox){

    			((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupEnteredAction(enteredAction,acrorend));
    		}else{

    		    currentComp.addMouseListener(formsActionHandler.setupEnteredAction(enteredAction,acrorend));
    		}
    	}

    	/*exited action*/
    	if(exitedAction!=null){
    		if(currentComp instanceof JComboBox){

    			((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupExitedAction(exitedAction,acrorend));
    		}else{
    		    currentComp.addMouseListener(formsActionHandler.setupExitedAction(exitedAction,acrorend));
    		}
    	}

    	/*activate action*/
    	if(activateAction!=null){
    	    if(activateAction.containsKey("ResetForm")){
    		    if(currentComp instanceof JComboBox){
    				((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupCommand("ResetForm",acrorend));
    			}else{
    			    currentComp.addMouseListener(formsActionHandler.setupCommand("ResetForm",acrorend));
    			}
    	    }

    	    if(activateAction.containsKey("Print")){
    		    if(currentComp instanceof JComboBox){
    				((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupCommand("Print",acrorend));
    			}else{
    			    currentComp.addMouseListener(formsActionHandler.setupCommand("Print",acrorend));
    			}
    	    }

    	    if(activateAction.containsKey("SaveAs")){
    		    if(currentComp instanceof JComboBox){
    				((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupCommand("SaveAs",acrorend));
    			}else{
    			    currentComp.addMouseListener(formsActionHandler.setupCommand("SaveAs",acrorend));
    			}
    	    }

    	    if(activateAction.containsKey("Hide")){
    	        if(currentComp instanceof JComboBox){
    				((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupHideAction(((Map)activateAction.get("Hide")),acrorend));
    			}else{
    			    currentComp.addMouseListener(formsActionHandler.setupHideAction(((Map)activateAction.get("Hide")),acrorend));
    			}
    	    }

    	    if(activateAction.containsKey("SubmitForm")){
    	    	Map aDataMap = (Map) activateAction.get("SubmitForm");

            	if(currentComp instanceof JComboBox){
        			((JComboBox)currentComp).getComponent(0).addMouseListener(formsActionHandler.setupSubmitAction(aDataMap,acrorend));
        		}else{
        		    currentComp.addMouseListener(formsActionHandler.setupSubmitAction(aDataMap,acrorend));
        		}
    	    }
    	}
    }

    /**
	 * setup the border style
	 */
	private  Border createBorderStyle(Object borderObj, Color borderColor,Color borderBackgroundColor) {
		/**Type must be Border
		 * W width in points (if 0 no border, default =1)
		 * S style - (default =S)
		 * 	S=solid, D=dashed (pattern specified by D entry below), B=beveled(embossed appears to above page),
		 * 	I=inset(engraved appeared to be below page), U=underline ( single line at bottom of boundingbox)
	     * D array phase - e.g. [a b] c means:-  a=on blocks,b=off blocks(if not present default to a),
	     * 		c=start of off block preseded index is on block.
	     * 	i.e. [4] 6 :- 4blocks on 4blocks off, block[6] if off - 1=off 2=on 3=on 4=on 5=on 6=off 7=off 8=off 9=off etc...
		 *
		 */
		if(printouts){
	        System.out.println("createBorderStyle() color="+borderColor+" background color="+borderBackgroundColor+"\n\tfield="+borderObj);
		}

		Map borderStream = new HashMap();
		if(borderObj==null){
		    borderStream.put("S","/S");
		    borderStream.put("W","1");
		}else{
		    if(borderObj instanceof Map){
		        borderStream = (Map)borderObj;
		    }else {
		    	LogWriter.writeFormLog("{DefaultFormFactory.createBorderStyle} border stream is String DefaultFormFactory.createBorderStyle",debugUnimplemented);
		    }
		}

		if(borderBackgroundColor==null){
//		    borderBackgroundColor = new Color(0,0,0,0);
		    if(printouts)
		        System.out.println("background border color null");
		}
		if(borderColor==null){
//		    borderColor = new Color(0,0,0,0);
		    if(printouts)
		        System.out.println("border color null");
		    return null;
		}

	    Border newBorder = null;

		String width = ((String)borderStream.get("W"));
	    if(printouts)
	        System.out.println("width="+width);

		String style = ((String)borderStream.get("S"));
	    style = Strip.checkRemoveLeadingSlach(style);
	    if(printouts)
	    	System.out.println("style="+style);

	    if(style.equals("U")){
	        if(printouts)
	            System.out.println("FormStream.createBorderStyle() U CHECK="+ConvertToString.convertMapToString(borderStream,null));
	        newBorder = BorderFactory.createMatteBorder(0,0,Integer.parseInt(width),0,borderColor);//underline field

	    }else if(style.equals("I")){
	        if(printouts)
            	System.out.println("FormStream.createBorderStyle() I CHECK="+ConvertToString.convertMapToString(borderStream,null));
            newBorder = BorderFactory.createEtchedBorder(borderColor,borderBackgroundColor);//inset below page

	    }else if(style.equals("B")){
	        if(printouts)
	            System.out.println("FormStream.createBorderStyle() B CHECK="+ConvertToString.convertMapToString(borderStream,null));
	        newBorder = BorderFactory.createBevelBorder(BevelBorder.LOWERED,borderColor,borderBackgroundColor);//beveled above page

	    }else if(style.equals("S")){
	        if(printouts)
	        	System.out.println("FormStream.createBorderStyle() S CHECK="+ConvertToString.convertMapToString(borderStream,null));
	        newBorder = BorderFactory.createLineBorder(borderColor,Integer.parseInt(width));//solid

	    }else if(style.equals("D")){
			Object dashPattern = borderStream.get("D");

	        if(dashPattern instanceof String){
	            if(debugUnimplemented)
					System.out.println("createBorderStyle D pattern is String="+(dashPattern));
	        }else if(dashPattern instanceof Map){
	            if(debugUnimplemented)
					System.out.println("createBorderStyle D pattern is Map="+ConvertToString.convertMapToString(((Map)dashPattern),null));
	        }else{
	            if(debugUnimplemented)
					System.out.println("createBorderStyle D pattern is UNKNOWN="+dashPattern);
	        }

			//setup dash pattern TODO MARK dashline
//	        new DashedBorder(Color.magenta,width,height);
//	    (new BasicStroke(3/scaling, BasicStroke.CAP_ROUND,
//	            BasicStroke.JOIN_ROUND, 0, new float[]{0,6/scaling,0,6/scaling}, 0));

	    }else{
	        if(debugUnimplemented)
			    System.out.println("NOT IMPLEMENTED BS - border stream="+ConvertToString.convertMapToString(borderStream,null));
	    }

	    return newBorder;
	}

	/**
	 * setup annotations display with popups, etc
	 */
	public Component annotationButton(final FormObject formObject) {
		//System.out.println("formobject="+formObject);

		JButton but = new JButton();//formObject.contents);
//		System.out.println("annotation text="+formObject.contents);
		but.setBorderPainted(false);
		but.setBorder(null);
		but.setOpaque(false);
		but.setContentAreaFilled(false);
		
		boolean[] fieldCharacteristic=formObject.characteristic;
	    Map enteredAction = formObject.enteredAction;
	    Map exitedAction = formObject.exitedAction;
	    Map activateAction = formObject.activateAction;
	    
       	setupMouseListener(but,fieldCharacteristic,enteredAction,exitedAction,activateAction);

		but.setIcon(new FixImageIcon(formObject.normalOffImage));
		but.setPressedIcon(new FixImageIcon(formObject.downOffImage));
		if(formObject.activateAction!=null){
			but.addMouseListener(formsActionHandler.setupClickedAction(formObject.activateAction,acrorend));
		}

		/**
		final JLabel url=new JLabel("<html><center>"+Messages.getMessage("PdfViewerJpedalLibrary.Text")
				+Messages.getMessage("PdfViewer.WebAddress"));
		url.setForeground(Color.blue);
		url.setHorizontalAlignment(JLabel.CENTER);
		url.setAlignmentX(Component.CENTER_ALIGNMENT);

		url.addMouseListener(new MouseListener() {
			public void mouseEntered(MouseEvent e) {
				details.setCursor(new Cursor(Cursor.HAND_CURSOR));
				url.setText("<html><center>"+Messages.getMessage("PdfViewerJpedalLibrary.Link")+
							Messages.getMessage("PdfViewerJpedalLibrary.Text")+
							Messages.getMessage("PdfViewer.WebAddress")+"</a></center>");
			}

			public void mouseExited(MouseEvent e) {
				details.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
				url.setText("<html><center>"+Messages.getMessage("PdfViewerJpedalLibrary.Text")
							+Messages.getMessage("PdfViewer.WebAddress"));
			}

			public void mouseClicked(MouseEvent e) {
				try {
					BrowserLauncher.openURL(Messages.getMessage("PdfViewer.VisitWebsite"));
				} catch (IOException e1) {
					showMessageDialog(Messages.getMessage("PdfViewer.ErrorWebsite"));
				}
			}

			public void mousePressed(MouseEvent e) {}
			public void mouseReleased(MouseEvent e) {}
		});*/

//		System.out.println("returning="+but);
		return but;
	}

    /**
     * setup and return the CheckBox button specified in the FormObject
     */
    public Component checkBoxBut(final FormObject form) {
        //			the text value
        JCheckBox checkBut = new JCheckBox(form.normalCaption);
        checkBut.setOpaque(false);
        
       	checkBut.setFont(form.textFont);
       	checkBut.setForeground(form.textColor);
        
        boolean[] fieldCharacteristic=form.characteristic;
	    Map enteredAction = form.enteredAction;
	    Map exitedAction = form.exitedAction;
	    Map activateAction = form.activateAction;
    	
    	setupMouseListener(checkBut, fieldCharacteristic,enteredAction,exitedAction,activateAction);
        
        if(form.appearancesUsed){
            
            appearanceImages(form,checkBut,showIconImages);
        	
       }else{
           Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
           checkBut.setBorder(newBorder);
           if(printouts){
           	System.out.println("borderStyle="+newBorder);
           }
           
           if(form.backgroundColor!=null){
               checkBut.setBackground(form.backgroundColor);
               checkBut.setOpaque(true);
           }
       }
        
        if(form.textPosition!=-1){
            /*
             * if there are any appearance images, then the text is set back to null,
             * if the textPosition needs to be setup you need to either set the text back here or not 
             * set it to null in appearanceImages. 
             * 
             * If you need to set this up check file acodabb.pdf page 4 as it has an icon with text being
             * set to overlay the icon, which doesn't work.
             */
            
            switch(form.textPosition){
            case 0: checkBut.setIcon(null);break;//0=caption only
            case 1: checkBut.setText(null);break;//1=icon only
            case 2: checkBut.setVerticalTextPosition(SwingConstants.BOTTOM);break;//2=caption below icon
            case 3: checkBut.setVerticalTextPosition(SwingConstants.TOP);break;//3=caption above icon
            case 4: checkBut.setHorizontalTextPosition(SwingConstants.RIGHT);break;//4=caption on right of icon
            case 5: checkBut.setHorizontalTextPosition(SwingConstants.LEFT);break;//5=caption on left of icon
            case 6: checkBut.setText(null);/*checkBut.setVerticalTextPosition(SwingConstants.CENTER);
            	checkBut.setHorizontalTextPosition(SwingConstants.CENTER);//6=caption overlaid ontop of icon */
            	break;
            }
        }
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            checkBut.setEnabled(false);
            checkBut.setDisabledIcon(checkBut.getIcon());
            checkBut.setDisabledSelectedIcon(checkBut.getSelectedIcon());
//              retComponent.setForeground(Color.magenta);
//              retComponent.setBackground(Color.magenta);
                if(debugUnimplemented)
                    System.out.println("READONLY comp="+checkBut);
            }
        
        return checkBut;
    }


    /**
     * setup and return the List field specified in the FormObject
     */
    public Component listField(final FormObject formObject) {
        if(printouts){
        	System.out.println("list box");
        }
        
        //populate the items array with list from Opt
        String[] items=formObject.itemsList;
        
        //create list (note we catch null value)
        JList list;
        if(items!=null)
        	list = new JList(items);
        else
        	list = new JList();
        
        //if there is a top index or selected value select it
        int topIndex = formObject.topIndex;
        if(topIndex!=-1){
        	String item = items[topIndex];
        	list.setSelectedValue(item,true);
        	if(printouts){
        		System.out.println("topIndex should be="+topIndex);
        	}
        }else {
        	String textValue = formObject.selectedItem;
        	list.setSelectedValue(textValue,true);
        	if(printouts)
        		System.out.println("currently selected value="+textValue);
        }
        
        
       	list.setFont(formObject.textFont);
       	list.setForeground(formObject.textColor);
        
        Border newBorder = createBorderStyle(formObject.border,formObject.borderColor,Color.white);
        list.setBorder(newBorder);
        if(printouts){
        	System.out.println("borderStyle="+newBorder);
        }
        
        if(formObject.backgroundColor!=null){
            list.setBackground(formObject.backgroundColor);
            list.setOpaque(true);
        }
        
        boolean[] fieldCharacteristic=formObject.characteristic;
        Map enteredAction = formObject.enteredAction;
        Map exitedAction = formObject.exitedAction;
        Map activateAction = formObject.activateAction;
        
        setupMouseListener(list,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }
        
        return list;
    }


    /**
     * setup and return the multi line password field specified in the FormObject
     */
    public Component multiLinePassword(final FormObject form) {
        JPasswordField multipass;
        
        int maxLength = form.maxTextLength;
        String textValue = form.textString;
        
        if(maxLength != -1)
        	multipass = new JPasswordField(textValue,maxLength);
        else
        	multipass = new JPasswordField(textValue);
        multipass.setEchoChar('*');
        multipass.setOpaque(false);
        
        if(printouts){
        	System.out.println("there is a password to be entered that can be MULTILINED");
        }
        
        Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
        multipass.setBorder(newBorder);
        if(printouts){
        	System.out.println("borderStyle="+newBorder);
        }
        
        if(form.backgroundColor!=null){
            multipass.setBackground(form.backgroundColor);
            multipass.setOpaque(true);
        }
        
       	multipass.setFont(form.textFont);
       	multipass.setForeground(form.textColor);
        
       	boolean[] fieldCharacteristic=form.characteristic;
	    Map enteredAction = form.enteredAction;
	    Map exitedAction = form.exitedAction;
	    Map activateAction = form.activateAction;
        
        setupMouseListener(multipass,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }
        
        //set text field alignment
        setAlignment(form.allignment, multipass);
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            multipass.setEditable(false);
//          retComponent.setForeground(Color.magenta);
//          retComponent.setBackground(Color.magenta);
            if(printouts)
                System.out.println("READONLY="+multipass);
        }
        
        return multipass;
    }


    /**
     * setup and return the multi line text area specified in the FormObject
     */
    public Component multiLineText(final FormObject form) {
        if(printouts)
        	System.out.println("MULTILINED no password");
        
        JTextArea newTextarea = new JTextArea(form.textString);
        newTextarea.setOpaque(false);
        
       	newTextarea.setFont(form.textFont);
       	newTextarea.setForeground(form.textColor);

        newTextarea.setLineWrap(true);

        //set length
    	int maxLength = form.maxTextLength;
        if(maxLength!=-1){
    		newTextarea.setColumns(maxLength);
    		if(printouts)
    			System.out.println("textlength added="+maxLength);
    	}
        
        Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
        newTextarea.setBorder(newBorder);
        if(printouts){
        	System.out.println("borderStyle="+newBorder);
        }
        
        if(form.backgroundColor!=null){
            newTextarea.setBackground(form.backgroundColor);
            newTextarea.setOpaque(true);
        }
        
        boolean[] fieldCharacteristic=form.characteristic;
	    Map enteredAction = form.enteredAction;
	    Map exitedAction = form.exitedAction;
	    Map activateAction = form.activateAction;
        
        setupMouseListener(newTextarea,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }

        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            newTextarea.setEditable(false);
//          retComponent.setForeground(Color.magenta);
//          retComponent.setBackground(Color.magenta);
            if(printouts)
                System.out.println("READONLY="+newTextarea);
        }
        
        return newTextarea;
    }
    
    /**
     * setup and return a signature field component,
     * <b>Note:</b> SKELETON METHOD FOR FUTURE UPGRADES.
     */
    public Component signature(FormObject formObject){
		
		JButton sigBut=new JButton(formObject.normalCaption);
        sigBut.setOpaque(false);
        
       	sigBut.setFont(formObject.textFont);
       	sigBut.setForeground(formObject.textColor);
       	
       	if(formObject.appearancesUsed){
            
            appearanceImages(formObject, sigBut,showIconImages);
            sigBut.setBorderPainted(false);
        }else{
            Border border = createBorderStyle(formObject.border,formObject.borderColor,Color.white);
            sigBut.setBorder(border);
            if(printouts){
            	System.out.println("border="+border);
            }
            
            if(formObject.backgroundColor!=null){
                sigBut.setBackground(formObject.backgroundColor);
                sigBut.setOpaque(true);
            }
        }
       	
       	Insets insetZero = new Insets(0,0,0,0);
        sigBut.setMargin(insetZero);
        
        if((formObject.flags!=null)&&(formObject.flags[1])){//FormStream.READONLY
        	sigBut.setEnabled(false);
        	sigBut.setDisabledIcon(sigBut.getIcon());
        	sigBut.setDisabledSelectedIcon(sigBut.getSelectedIcon());
//      	retComponent.setForeground(Color.magenta);
//      	retComponent.setBackground(Color.magenta);
        	if(debugUnimplemented)
        		System.out.println("READONLY comp="+sigBut);
        }
        
        return sigBut;
    }
    
    /**
     * setup and return the Push button specified in the FormObject
     */
    public Component pushBut(final FormObject form) {
        //			the text value
        JButton pushBut=new JButton(form.normalCaption);
        pushBut.setOpaque(false);
        pushBut.setContentAreaFilled(false);
        
       	pushBut.setFont(form.textFont);
       	pushBut.setForeground(form.textColor);
        
       	boolean[] fieldCharacteristic=form.characteristic;
	    Map enteredAction = form.enteredAction;
	    Map exitedAction = form.exitedAction;
	    Map activateAction = form.activateAction;
	    
       	setupMouseListener(pushBut,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        
        if(form.appearancesUsed){
            
            appearanceImages(form, pushBut,showIconImages);
        }else{
            Border border = createBorderStyle(form.border,form.borderColor,Color.white);
            pushBut.setBorder(border);
            if(printouts){
            	System.out.println("border="+border);
            }
            
            if(form.backgroundColor!=null){
                pushBut.setBackground(form.backgroundColor);
                pushBut.setOpaque(true);
            }
        }
        
        if(form.appearancesUsed){
            /*
             * if there are any appearance images, then the text is set back to null,
             * if the textPosition needs to be setup you need to either set the text back here or not 
             * set it to null in appearanceImages. 
             * 
             * If you need to set this up check file acodabb.pdf page 4 as it has an icon with text being
             * set to overlay the icon, which doesn't work.
             */
            
            switch(form.textPosition){
            case 0: pushBut.setIcon(null);break;//0=caption only
            case 1: pushBut.setText(null);break;//1=icon only
            case 2: pushBut.setVerticalTextPosition(SwingConstants.BOTTOM);break;//2=caption below icon
            case 3: pushBut.setVerticalTextPosition(SwingConstants.TOP);break;//3=caption above icon
            case 4: pushBut.setHorizontalTextPosition(SwingConstants.RIGHT);break;//4=caption on right of icon
            case 5: pushBut.setHorizontalTextPosition(SwingConstants.LEFT);break;//5=caption on left of icon
            case 6: pushBut.setText(null);/*pushBut.setVerticalTextPosition(SwingConstants.CENTER);
            	pushBut.setHorizontalTextPosition(SwingConstants.CENTER);//6=caption overlaid ontop of icon */
            	break;
            }
        }
        
        Insets insetZero = new Insets(0,0,0,0);
        pushBut.setMargin(insetZero);
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            pushBut.setEnabled(false);
            pushBut.setDisabledIcon(pushBut.getIcon());
            pushBut.setDisabledSelectedIcon(pushBut.getSelectedIcon());
//          retComponent.setForeground(Color.magenta);
//          retComponent.setBackground(Color.magenta);
            if(debugUnimplemented)
                System.out.println("READONLY comp="+pushBut);
        }
        
        return pushBut;
    }


    /**
     * setup and return the Radio button specified in the FormObject
     */
    public Component radioBut(final FormObject form) {
        
        //the text value
        JRadioButton radioBut = new JRadioButton(form.normalCaption);
        radioBut.setOpaque(false);
    	//radioBut.setContentAreaFilled(false);//false for transparency
    
       	radioBut.setFont(form.textFont);
       	radioBut.setForeground(form.textColor);
        
       	boolean[] fieldCharacteristic=form.characteristic;
	    Map enteredAction = form.enteredAction;
	    Map exitedAction = form.exitedAction;
	    Map activateAction = form.activateAction;
        
        setupMouseListener(radioBut,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }
        
        if(form.appearancesUsed){
           
        	appearanceImages(form, radioBut,showIconImages);
        
        }else{
            Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
            radioBut.setBorder(newBorder);
            if(printouts){
            	System.out.println("borderStyle="+newBorder);
            }
            
            if(form.backgroundColor!=null){
                radioBut.setBackground(form.backgroundColor);
                radioBut.setOpaque(true);
            }
        }
        
        if(form.textPosition!=-1){
            /*
             * if there are any appearance images, then the text is set back to null,
             * if the textPosition needs to be setup you need to either set the text back here or not 
             * set it to null in appearanceImages. 
             * 
             * If you need to set this up check file acodabb.pdf page 4 as it has an icon with text being
             * set to overlay the icon, which doesn't work.
             */
            
            switch(form.textPosition){
            case 0: radioBut.setIcon(null);break;//0=caption only
            case 1: radioBut.setText(null);break;//1=icon only
            case 2: radioBut.setVerticalTextPosition(SwingConstants.BOTTOM);break;//2=caption below icon
            case 3: radioBut.setVerticalTextPosition(SwingConstants.TOP);break;//3=caption above icon
            case 4: radioBut.setHorizontalTextPosition(SwingConstants.RIGHT);break;//4=caption on right of icon
            case 5: radioBut.setHorizontalTextPosition(SwingConstants.LEFT);break;//5=caption on left of icon
            case 6: radioBut.setText(null);/*radioBut.setVerticalTextPosition(SwingConstants.CENTER);
            	radioBut.setHorizontalTextPosition(SwingConstants.CENTER);//6=caption overlaid ontop of icon */
            	break;
            }
        }
        
        Insets insetZero = new Insets(0,0,0,0);
        radioBut.setMargin(insetZero);
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            radioBut.setEnabled(false);
            radioBut.setDisabledIcon(radioBut.getIcon());
            radioBut.setDisabledSelectedIcon(radioBut.getSelectedIcon());
//              retComponent.setForeground(Color.magenta);
//              retComponent.setBackground(Color.magenta);
                if(debugUnimplemented)
                    System.out.println("READONLY comp="+radioBut);
            }
        
        return radioBut;
    }


    /**
     * setup and return the single line password field specified in the FormObject
     */
    public Component singleLinePassword(final FormObject form) {
        if(printouts)
        	System.out.println("there is a password to be entered single");
        
        JPasswordField newPassword=new JPasswordField(form.textString);
        newPassword.setEchoChar('*');
        newPassword.setOpaque(false);
        
       	newPassword.setFont(form.textFont);
       	newPassword.setForeground(form.textColor);
        
        //				set length
       	int maxLength = form.maxTextLength;
       	if(maxLength!=-1){
       		newPassword.setColumns(maxLength);
       		if(printouts)
       			System.out.println("textlength added="+maxLength);
       	}
        
        Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
        newPassword.setBorder(newBorder);
        if(printouts){
        	System.out.println("borderStyle="+newBorder);
        }
        
        if(form.backgroundColor!=null){
            newPassword.setBackground(form.backgroundColor);
            newPassword.setOpaque(true);
        }
        
        boolean[] fieldCharacteristic=form.characteristic;
	    Map enteredAction = form.enteredAction;
	    Map exitedAction = form.exitedAction;
	    Map activateAction = form.activateAction;
        
        setupMouseListener(newPassword,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }
        
        //set text field alignment
        setAlignment(form.allignment, newPassword);
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            newPassword.setEditable(false);
//          retComponent.setForeground(Color.magenta);
//          retComponent.setBackground(Color.magenta);
            if(printouts)
                System.out.println("READONLY="+newPassword);
        }
        
        return newPassword;
    }


    /**
     * setup and return the single line text field specified in the FormObject
     */
    public Component singleLineText(final FormObject form) {
        if(printouts)
        	System.out.println("no password singleline");
        
        JTextField newTextfield = new JTextField(form.textString);
        newTextfield.setOpaque(false);
        
        if(form.textFont!=null)
       	newTextfield.setFont(form.textFont);
       	    
       	newTextfield.setForeground(form.textColor);
        
        //set length
    	int maxLength = form.maxTextLength;
    	if(maxLength!=-1){
    		newTextfield.setColumns(maxLength);
    		if(printouts)
    			System.out.println("textlength added="+maxLength);
    	}
        
        Border newBorder = createBorderStyle(form.border,form.borderColor,Color.white);
        newTextfield.setBorder(newBorder);
        if(printouts){
        	System.out.println("borderStyle="+newBorder);
        }
        
        if(form.backgroundColor!=null){
            newTextfield.setBackground(form.backgroundColor);
            newTextfield.setOpaque(true);
        }
        
        boolean[] fieldCharacteristic=form.characteristic;
        Map enteredAction = form.enteredAction;
        Map exitedAction = form.exitedAction;
        Map activateAction = form.activateAction;
        
        setupMouseListener(newTextfield,fieldCharacteristic,enteredAction,exitedAction,activateAction);
        
        //set text field alignment
        setAlignment(form.allignment, newTextfield);
        
        if(printouts){
        	System.out.println("flagNum="+fieldCharacteristic);
        }
        
        if((form.flags!=null)&&(form.flags[1])){//FormStream.READONLY
            newTextfield.setEditable(false);
//          retComponent.setForeground(Color.magenta);
//          retComponent.setBackground(Color.magenta);
            if(printouts)
                System.out.println("READONLY="+newTextfield);
        }
        
        return newTextfield;
    }


    /**
     * set the allignment of the field passed in
     */
    private void setAlignment(int align, JTextField newTextfield) {
    	if(align!=-1){
    		newTextfield.setHorizontalAlignment(align);
    	}
    }


    /**
     * gets each appearance image from the map <b>appearance</b> and
     * and adds it to the relevent icon for the AbstractButton <b>comp</b>
     * showImages is to display the appearance images for that FormObject
     */
    private void appearanceImages(final FormObject form, AbstractButton comp,boolean showImages) {
        //get default state
        String defaultState = form.defaultState;
        if(printouts)
            System.out.println("default state="+defaultState);
        if(defaultState!=null && defaultState.equals("On")){
            comp.setSelected(true);
        }
        
        if(form.hasNormalOff()){
            comp.setText(null);
           	comp.setIcon(new FixImageIcon(form.normalOffImage));
            if(showImages)
                ShowGUIMessage.showGUIMessage("normalAppOffImage",form.normalOffImage,"normalAppOff");
        }
        if(form.hasNormalOn()){
            comp.setText(null);
            comp.setSelectedIcon(new FixImageIcon(form.normalOnImage));
            if(showImages)
                ShowGUIMessage.showGUIMessage("normalAppOnImage",form.normalOnImage,"normalAppOn");
        }
        
        if(form.noDownIcon){
            comp.setEnabled(false);
        }else {
            BufferedImage downOff = null,downOn = null;
            
            if(form.offsetDownIcon && !form.hasDownImage()){
                if(form.hasNormalOff()){
                    if(form.hasNormalOn()){
                        downOn = createPressedLook(form.normalOnImage);
                        downOff = createPressedLook(form.normalOffImage);
                    }else {
                        downOff = createPressedLook(form.normalOffImage);
                    }
                }else if(form.hasNormalOn()){
                    downOff = createPressedLook(form.normalOnImage);
                }
            }else if(form.invertDownIcon){
            	if(form.hasNormalOff()){
                    if(form.hasNormalOn()){
                    	downOff = invertImage(form.normalOffImage);   
        	    		downOn = invertImage(form.normalOnImage);
                    }else {
                    	downOff = invertImage(form.normalOffImage);
                    }
                }else if(form.hasNormalOn()){
                	downOff = invertImage(form.normalOnImage);
                }
            }else {
                downOff = form.downOffImage;
                downOn = form.downOnImage;
            }
            
            if(showImages){
                ShowGUIMessage.showGUIMessage("downAppOffImage",form.downOffImage,"downAppOff");
                ShowGUIMessage.showGUIMessage("downAppOnImage",form.downOnImage,"downAppOn");
            }
            
            if(downOff==null || downOn==null){
                
                if(downOff!=null){
                    comp.setText(null);
                    comp.setPressedIcon(new FixImageIcon(downOff));
                }else if(downOn!=null){
                    comp.setText(null);
                    comp.setPressedIcon(new FixImageIcon(downOn));
                }
                
                if(printouts)
                    System.out.println("CHECK there is an off or on down image not both does this work, DefaultAcro.createAppearanceImages");
            }else{
                
                if(comp.isSelected())
                    comp.setPressedIcon(new FixImageIcon(downOn));
                else
                    comp.setPressedIcon(new FixImageIcon(downOff));
                
                comp.addActionListener(formsActionHandler.setupChangingDownIcon(downOff,downOn));
            }
        }
        
        if(form.hasRolloverOff()){
            comp.setRolloverEnabled(true);
            comp.setText(null);
            comp.setRolloverIcon(new FixImageIcon(form.rolloverOffImage));
            if(showImages)
                ShowGUIMessage.showGUIMessage("rolloverAppOffImage",form.rolloverOffImage,"rolloverAppOff");
        }
        if(form.hasRolloverOn()){
            comp.setRolloverEnabled(true);
            comp.setText(null);
            comp.setRolloverSelectedIcon(new FixImageIcon(form.rolloverOnImage));
            if(showImages)
                ShowGUIMessage.showGUIMessage("rolloverAppOnImage",form.rolloverOnImage,"rolloverAppOn");
        }
    }
    
    private BufferedImage invertImage(BufferedImage image) {
    	BufferedImage ret = new BufferedImage(image.getWidth(),image.getHeight(),image.getType());
        
        byte reverse[] = new byte[256]; 
        for (int j=0; j<200; j++){  
            		reverse[j]=(byte)(256-j);  
     	}	 
        ByteLookupTable blut=new ByteLookupTable(0, reverse);  
        LookupOp lop = new LookupOp(blut, null);  
        lop.filter(image,ret);
        
        return ret;
	}

	/**
     * create a pressed look of the <b>image</b> and return it
     */
    private BufferedImage createPressedLook(Image image) {
    	BufferedImage pressedImage = new BufferedImage(image.getWidth(null)+2,image.getHeight(null)+2,BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = (Graphics2D)pressedImage.getGraphics();
        g.drawImage(image, 1, 1, null);
        g.dispose();
        return pressedImage;
    }

}
